"""
Agent loops for agent
"""

# Import the loops to register them
from . import anthropic
from . import openai
from . import uitars
from . import omniparser
from . import gta1
from . import composed_grounded
from . import glm45v
from . import opencua
from . import internvl
from . import holo
from . import moondream3
from . import gemini

__all__ = [
    "anthropic", 
    "openai", 
    "uitars", 
    "omniparser", 
    "gta1", 
    "composed_grounded", 
    "glm45v", 
    "opencua",
    "internvl",
    "holo",
    "moondream3",
    "gemini"
]